// *******************************************************************************
// File name:		algorthm.cpp
// Author:		Laura Dekker
// Description:		Algorithm - base algorithm class
//                      
//
// Change history:
//
//
// *******************************************************************************

#include <string.h>
#include "evaluate.h"
#include "pools.h"
#include "algorthm.h"



/*****************************************************************
 Class Algorithm

 Base class for any algorithm created with this library
 *****************************************************************/




/*****************************************************************
 Define the static evaluator pointer
 *****************************************************************/

Evaluator* Algorithm::_evaluator;



/*****************************************************************
 Constructor and destructor
 *****************************************************************/

Algorithm::Algorithm(void)
{
	for (int p=0; p<MAX_POOLS; p++)        
		_pool_array[p] = 0;
        

	_evaluator = 0;
	_optimise = MAXIMISE;	//default
	_iteration = 0;
	
	strcpy(_output_file, "");
	_output_detail = BEST_ONLY;
	_output_frequency = 2;
} 


Algorithm::~Algorithm(void)
{
	for (int p=0; p<MAX_POOLS; p++)
        {
		delete _pool_array[p];
		_pool_array[p] = 0;
	} 	
       
	output.close();
	delete _evaluator;	
}



/*****************************************************************
 Pool management
 *****************************************************************/

void Algorithm::SetPool(int index, Pool* pool)
{
	if (index<0  ||  index>=MAX_POOLS)
        	return;

	delete _pool_array[index];
	_pool_array[index] = pool;
}


Pool* Algorithm::GetPool(int index)
{
	if (index<0  ||  index>=MAX_POOLS)
		return 0;
	else
		return _pool_array[index];
}


int Algorithm::GetNumPools(void)
{	return _num_pools;	}


void Algorithm::SetNumPools(int num_pools)
{	_num_pools = num_pools;	}




/*****************************************************************
 Start-up operations before the algorithm run
 *****************************************************************/

void Algorithm::StartGA(void)
{	_evaluator->StartGA();	}



/*****************************************************************
 Run the algorithm
 *****************************************************************/

void Algorithm::Run(void)
{
	StartGA();
	while (!EndConditions())
		Iteration();
	EndGA();
}


/********************************************************************
 Do post-processing when the run has ended
********************************************************************/

void Algorithm::EndGA(void)
{	_evaluator->EndGA();	}




/********************************************************************
 Get and set iteration counter and max iterations
********************************************************************/

long Algorithm::GetIteration(void)
{	return _iteration;	}

void Algorithm::SetIteration(long iteration)
{	_iteration = iteration;	}

long Algorithm::GetMaxIterations(void)
{	return _max_iterations;	}

void Algorithm::SetMaxIterations(long max_iterations)
{	_max_iterations = max_iterations;	}




/*****************************************************************
 Evaluator access functions
 *****************************************************************/

void Algorithm::SetEvaluator(Evaluator* evaluator)
{
	delete _evaluator;
	_evaluator = evaluator;
}


Evaluator* Algorithm::GetEvaluator(void)
{       return _evaluator;      }



void Algorithm::SetEvaluatorType(EvaluatorType type)
{	_evaluator_type = type;	}


EvaluatorType Algorithm::GetEvaluatorType(void)
{	return _evaluator_type;	}



/*****************************************************************
 Optimise by maximising or minimising
 *****************************************************************/

void Algorithm::SetOptimiseType(OptimiseType min_or_max)
{	_optimise = min_or_max;	 }

OptimiseType Algorithm::GetOptimiseType(void) const
{	return _optimise;	}




Individual* Algorithm::GetBestIndividual(void)
{
	Individual* best_individual = _pool_array[0]->GetOverallBest();
	FitnessType best_fitness = best_individual->GetFitness();

	for (int p=1; p<_num_pools; p++)
	{         	
		if ((GetOptimiseType() == MINIMISE  &&  _pool_array[p]->GetOverallBest()->GetFitness() < best_fitness)
			|| (GetOptimiseType() == MAXIMISE  &&  _pool_array[p]->GetOverallBest()->GetFitness() > best_fitness))

                {
			best_individual = _pool_array[p]->GetOverallBest();
			best_fitness = best_individual->GetFitness();
		}
	}
	
	return best_individual;
}



/*****************************************************************
 Functions for outputting algorithm run statistics
 *****************************************************************/

void Algorithm::SetOutputFile(char* path)
{
	if (strcmp(path, _output_file))
        {
		strncpy(_output_file, path, MAX_FILE_PATH-1);
		output.close();
		output.open(_output_file);
	}
}

char* Algorithm::GetOutputFile(void)
{	return _output_file;	}



void Algorithm::SetOutputDetail(StatsLevel detail)
{	_output_detail = detail;	}

StatsLevel Algorithm::GetOutputDetail(void)
{	return _output_detail;	}

void Algorithm::SetOutputFrequency(int frequency)
{	_output_frequency = frequency;	}

int Algorithm::GetOutputFrequency(void)
{	return _output_frequency;	}


