// W.Langdon@cs.ucl.ac.uk 5-dec-94

#include "pools.h"
#include "individ.h"
#include "grideval.h"
#include "gridalg.h"


GridGA::GridGA(void)
{
	// define pools to have default settings (but don't create yet)
	_num_pools = DEF_NUM_POOLS;
	_num_individuals = DEF_POPULATION_SIZE;
	_prototype = new Individual(DEF_NUM_CHROMOSOMES, DEF_CHROMOSOME_LENGTH, DEF_GENE_TYPE);
	
	_initialisation = 0;
	_crossover = 0;
	_mutation = 0;
	_selection = 0;
	_migration = 0;

	strcpy(_problem_type, "combinatorial");

	_extraAttributes = 0;

	_iteration = 0;
	_max_iterations = DEF_MAX_GEN;

	// create the evaluator
	GridEval* evaluator = new GridEval("dummy filename");
	_evaluator_type = E_DLL; //close enough
	SetEvaluator(evaluator);
}


GridGA::~GridGA(void)
{
	ResetGA();
	delete _prototype;
	delete _evaluator;
	
	delete _initialisation;
	delete _crossover;
	delete _mutation;
	delete _selection;
	delete _migration;
}




/*****************************************************************
 Algorithm run functions
 *****************************************************************/


/*****************************************************************
 StartUp - called before the first iteration
 *****************************************************************/

void GridGA::StartUp(void)
{
        SetOptimiseType(MINIMISE);

	if (!strcmp(_problem_type, "normal")) // standard value-based problem
        {
		_initialisation = new RandomInitialise;
		_crossover = new HollandCrossover;
		_mutation = new BasicMutation;
	}
	else // combinatorial problem	
        {
		_initialisation = new CombinatorialInitialise;
		_crossover = new CombinatorialCrossover;
		_mutation = new CombinatorialMutation;
	}
	_selection = new TruncatedSelection;
//	_migration = new BestToWorstMigration;

	// create the pools
	for (int p=0; p<_num_pools; p++)
	{	
		_pool_array[p] = new Pool(_num_individuals, _prototype);	
		_initialisation->DoInitialisation(_pool_array[p]);
		_pool_array[p]->StoreOverallBest(_optimise);
		_pool_array[p]->SetOldAvgFitness(_pool_array[p]->GetAverageFitness());
	}
}



/*****************************************************************
 Iteration - called each time through the iteration loop
 *****************************************************************/

void GridGA::Iteration(void)
{
	_iteration++;

	for (int p=0; p<_num_pools; p++)
	{
		_pool_array[p]->SetOldAvgFitness(_pool_array[p]->GetAverageFitness());

		#ifdef DEBUG1
		output << "Generation: " << _iteration << "  Parents Pool " << p << '\n';
		_pool_array[p]->Statistics(output);
		output <<"Average fitness = " << _pool_array[p]->GetAverageFitness() << '\n';
		#endif

		if (_num_pools > 1)
		{
			#ifdef DEBUG1
			output << "Doing migration\n";
			#endif
			for (int i = 0; i<_num_pools; i++)
                        {
				if (i != p)
					_migration->DoMigration(_pool_array[p], _pool_array[i], _optimise);
			}
		}

		#ifdef DEBUG1
		output << "Doing crossover\n";
		#endif
		_crossover->DoCrossover(_pool_array[p]);
		#ifdef DEBUG1
		output << "Doing mutation\n";
		#endif
		_mutation->DoMutation(_pool_array[p]);

		#ifdef DEBUG1
		output << "Generation: " << _iteration << "  Offspring Pool " << p << '\n';
		_pool_array[p]->Statistics(output);
		output <<"Average fitness = " << _pool_array[p]->GetAverageFitness() << '\n';
		#endif

		// if one of the current pool is the highest so far, save it
		_pool_array[p]->StoreOverallBest(_optimise);

		#ifdef DEBUG1		
		output << "Doing selection\n";
		#endif
		_selection->DoSelection(_pool_array[p], _optimise);

		#ifdef DEBUG1
		output << "Generation: " << _iteration << "  Mating Pool " << p << '\n';
		_pool_array[p]->Statistics(output);
		output <<"Average fitness = " << _pool_array[p]->GetAverageFitness() << '\n';
		#endif
	} 	
}



/********************************************************************
 Check for end conditions
 return TRUE if conditions met, return FALSE if not
********************************************************************/

int GridGA::EndConditions(void)
{
	if (_iteration < _max_iterations)
		return FALSE;
	else
		return TRUE;
}


/*****************************************************************
 Return existing algorithm to its original state, but don't lose
 the settings
 *****************************************************************/

void GridGA::ResetGA(void)
{
	if (_pool_array[0])
        {
		for (int pool=0; pool<_num_pools; pool++)
                {
			delete _pool_array[pool];
			_pool_array[pool] = 0;
                }
	}

	_iteration = 0;
}






/********************************************************************
 Pool set-up functions
********************************************************************/

long GridGA::GetNumIndividuals(void)
{	return _num_individuals;	}
 
void GridGA::SetNumIndividuals(long num_individuals)
{	_num_individuals = num_individuals;	}


Individual* GridGA::GetPrototype(void)
{	return _prototype;	}

void GridGA::SetPrototype(Individual* prototype)
{
	//delete _prototype;
	_prototype = prototype;
}

