// *******************************************************************************
// File name:		individ.cpp
// Author:		Laura Dekker
// Description:		Classes for Gene, Chromosome and Individual
//
//
// Change history:
//
//
// *******************************************************************************

#include <math.h>
#include "individ.h"
#include "algorthm.h"
#include "evaluate.h"

#define DEBUG


/*****************************************************************
 Class Gene

 Used only where a chromosome has genes with mixed ranges;
 otherwise the primitive types are used: char, int, long, double
 *****************************************************************/

Gene::Gene(void)
{
	_min = 0;
	_max = 0;
	_value = 0;
}


Gene::Gene(const Gene& old_gene)
{
	*this = old_gene; 
}


Gene& Gene::operator=(const Gene& rhs)
{
	_value = rhs._value;
	_min = rhs._min;
	_max = rhs._max;   
	return *this;
}


Gene& Gene::operator=(const double value)
{
	_value = value;
	return *this;
}


Gene::~Gene(void)
{}


/*****************************************************************
 casts
 *****************************************************************/

Gene::operator int()
{       return (int)_value;     }


Gene::operator long()
{       return (long)_value;    }


Gene::operator double()
{       return (double)_value;  }


/*****************************************************************
 get and set min and max values for gene
 *****************************************************************/

void Gene::SetMin(double min)
{       _min = min;     }


double Gene::GetMin(void)
{       return _min;    }


void Gene::SetMax(double max)
{       _max = max;     }


double Gene::GetMax(void)
{       return _max;    }


/*****************************************************************
 iostreams for gene 
 *****************************************************************/

ostream& operator<< (ostream& os, const Gene& gene)
{
	os << gene._value;
	return os;
}



Gene& operator>> (istream& is, Gene& gene)
{
	double in_value;
	is >> in_value;
        
	if (in_value < gene._min)
		in_value = gene._min;
	else if (in_value > gene._max)
		in_value = gene._max;

	gene._value = in_value;
	return gene;
}






/*****************************************************************
 Class Chromosome
 contains an array of genes of one of several types
 
 can set gene value range over whole chromosome 
 *****************************************************************/

Chromosome::Chromosome(void)
{
	_num_genes = 0;
	_gene_array = 0;

	_min = 0;
	_max = MAX_CHAR;

	_type = G_INT;
	_alphabet = 0;
}



Chromosome::Chromosome(long num_genes, GeneType type)
{
	_num_genes = num_genes;	
	_min = 0;
	_type = type;

	if (_type == G_ENCODED)
	{
		_gene_array = (void*)(new char[_num_genes]);
		_alphabet = 2;	// binary by default;
		_max = pow(2,_num_genes)-1;
	}

	else
        {
		switch (_type)
		{ 
			case G_INT:
				_gene_array = (void*)(new long[_num_genes]);
				_max = MAX_CHAR;
				break;

			case G_FLOAT:
				_gene_array = (void*)(new double[_num_genes]);
				_max = MAX_CHAR;
				break;

			case G_USER:
				_gene_array = (void*)(new Gene[_num_genes]);
				break;
		}

		_alphabet = 0;
	}
}



Chromosome::Chromosome(long num_genes, char alphabet)
{
	_num_genes = num_genes;

	_type = G_ENCODED;
	_alphabet = alphabet;

	_gene_array = (void*)(new char[_num_genes]);

	_min = 0;
	_max = pow(_alphabet,_num_genes)-1;	// default value
}



Chromosome::Chromosome(const Chromosome& old_chromosome)
{
	_gene_array = 0;
	*this = old_chromosome;
}



// fitness invalid
Chromosome& Chromosome::operator=(const Chromosome& rhs)
{
	int g;

	_type = rhs._type;
	_alphabet = rhs._alphabet;
	_min = rhs._min;
	_max = rhs._max;
	_num_genes = rhs._num_genes;

	switch (_type)
	{		
		case G_ENCODED:
			delete[] (char*)_gene_array;
			_gene_array = (void*)(new char[_num_genes]);
			for (g=0; g<_num_genes; g++)
				((char*)_gene_array)[g] = ((char*)(rhs._gene_array))[g];
			break;

		case G_INT:
			delete[] (long*)_gene_array;
			_gene_array = (void*)(new long[_num_genes]);
			for (g=0; g<_num_genes; g++)
				((long*)_gene_array)[g] = ((long*)(rhs._gene_array))[g];
			break;

		case G_FLOAT:
			delete[] (double*)_gene_array;
			_gene_array = (void*)(new double[_num_genes]);
			for (g=0; g<_num_genes; g++)
				((double*)_gene_array)[g] = ((double*)(rhs._gene_array))[g];
			break;
			
		case G_USER:
			delete[] (Gene*)_gene_array;
			_gene_array = (void*)(new Gene[_num_genes]);
			for (g=0; g<_num_genes; g++)
				((Gene*)_gene_array)[g] = ((Gene*)(rhs._gene_array))[g];
			break;
	}

	return *this;
}


Chromosome::~Chromosome(void)
{
	switch (_type)
	{		
		case G_ENCODED:
			delete[] (char*)_gene_array;
			break;

		case G_INT:
			delete[] (long*)_gene_array;
			break;

		case G_FLOAT:
			delete[] (double*)_gene_array;
			break;
			
		case G_USER:
			delete[] (Gene*)_gene_array;
			break;
	}
}


/*****************************************************************
 gene access functions
 *****************************************************************/

double Chromosome::ReadGene(long index)
{
	double value;

	if (index <= _num_genes)
	{
		switch (_type)
		{ 			
			case G_ENCODED:
				value = ((char*)_gene_array)[index];
				break;

			case G_INT:
				value = ((long*)_gene_array)[index];
				break;

			case G_FLOAT:
				value = ((double*)_gene_array)[index];
				break;
				
			case G_USER:
				value = ((Gene*)_gene_array)[index];
				break;
		}
	}
	return value;
}


// fitness invalid
void Chromosome::WriteGene(long index, double value)
{
	if (index <= _num_genes)
	{
		if (_type == G_ENCODED)
                {
			((char*)_gene_array)[index] = value;
			// if we've just written a bit that takes the decoded
			// value out of range, truncate it				
			if (BaseDecode() > _max)
				BaseEncode(_max);
		}

		else
                {  
			// if value is out of bounds, truncate to min/max                
			if (value < _min)
				value = _min;
			else if (value > _max)
				value = _max;

			switch (_type)
			{
				case G_INT:
					((long*)_gene_array)[index] = value;
					break;

				case G_FLOAT:
					((double*)_gene_array)[index] = value;
					break;
				
				case G_USER:
					((Gene*)_gene_array)[index] = value;
					break;
			}
		}
	}
}



/*****************************************************************
 Get the alphabet for an encoded chromosome
 *****************************************************************/

char Chromosome::GetAlphabet(void)
{
	if (_type == G_ENCODED)
		return _alphabet;
	else
		return 0;
}



/*****************************************************************
 Encoding and decoding functions
 Max number of bits = MAX_LONG
 For fill to be 0, values must be unsigned
 *****************************************************************/

unsigned long Chromosome::BinaryDecode(void) const
{
	unsigned long shifted, value;

	for (char index = 0; index < _num_genes; index++)
	{
		if (index == 0)		
			shifted = value = 0;
		else
                {
        		// shift the value left
			shifted = value << 1;
		} 
		// put the next bit into the last exposed position		
		value = shifted | ((char*)_gene_array)[index];		
	}
	return value;
}


void Chromosome::BinaryEncode(unsigned long value)
{
	unsigned long shifted;

	for (char index = 0; index < _num_genes; index++)
	{
		// shift the value right to make the current bit the end one
		shifted = value >> index;
		// AND it with a mask on the final bit to get the bit value
		// Put the bit values into the gene array starting from the end		
		((char*)_gene_array)[(_num_genes-index)-1] = shifted & 1;
	}
	// throws away any remaining bits left in the value
}




/*****************************************************************
 general encoding for any base (assumes max value for chromosome
 is the base)

 if value given > max number for that base/num_genes, automatically
 sets to max

 makes fitness cache invalid
 *****************************************************************/

void Chromosome::BaseEncode(unsigned long value)
{
	if (value > _max)
		value = _max;

	if (_alphabet == 2)
		BinaryEncode(value);
	else
        {
		unsigned long real_val = value;	
		
		char exponent = _num_genes-1;
		for (char index=0; index<_num_genes; index++, exponent--)
		{
			double tmp = pow(_alphabet,exponent);

			for(int al_set=_alphabet-1; al_set>=0; al_set--)
			{
				((char*)_gene_array)[index] = al_set;
				if (real_val - tmp*(double)al_set >= 0)
                        	{
					real_val = real_val-tmp*(double)al_set;
                        		break;
				}
			}
		}
	}

	return; 
}



/*****************************************************************
 general decoding for any base (assumes max value for chromosome
 is the base)
 *****************************************************************/

unsigned long Chromosome::BaseDecode(void) const
{
	unsigned long real_val = 0;

	if (_alphabet == 2)
		real_val = BinaryDecode();
	else
        {
		char exponent = _num_genes-1;
		for (char index=0; index<_num_genes; index++, exponent--)
		{  		
			char al_set = ((char*)_gene_array)[index];
			real_val += al_set*pow(_alphabet,exponent);
		}
	}

	return real_val;
}



/*****************************************************************
 Convert an encoded chromosome to a new base
 *****************************************************************/

void Chromosome::ConvertBase(char new_alphabet)
{
	if (new_alphabet == _alphabet)
		return;

	// store the existing decoded value
	long chrom_value = BaseDecode();

	// calculate the number of chromosomes needed for the new base
	double num_genes = log(_max)/log(new_alphabet);
	if (num_genes > (int)num_genes)
		num_genes = (int)num_genes + 1; 

	// delete the old gene array
	if (num_genes != _num_genes)
	{
		_num_genes = num_genes;
		delete[] (char*)_gene_array;
		_gene_array = (void*)(new char[_num_genes]);
	}

	_alphabet = new_alphabet;
	BaseEncode(chrom_value);
}



/*****************************************************************
 get and set gene min and max values
 *****************************************************************/

void Chromosome::SetMinValue(double min)
{       _min = min;	}


double Chromosome::GetMinValue(void)
{       return _min;	}


void Chromosome::SetMaxValue(double max)
{       _max = max;	}


double Chromosome::GetMaxValue(void)
{      	return _max;	}



/*****************************************************************
 iostreams
 *****************************************************************/

ostream& operator<< (ostream& os, const Chromosome& chromosome)
{
	if (chromosome._type == G_ENCODED)
			os << chromosome.BaseDecode() << '\t';

	else for (long g=0; g<chromosome._num_genes; g++)
	{
		switch (chromosome._type)
		{
			case G_INT:
				os << ((long*)(chromosome._gene_array))[g] << '\t';
				break;

			case G_FLOAT:
				os << ((double*)(chromosome._gene_array))[g] << '\t';
				break;

			case G_USER:
				// this will output all Genes as doubles...hmmm...
				os << (double)(((Gene*)(chromosome._gene_array))[g]) << '\t';
				break;
		}
	}
	return os;
}



Chromosome& operator>> (istream& is, Chromosome& chromosome)
{
	if (chromosome._type == G_ENCODED)
	{
		unsigned long in_value;
		is >> in_value;
		chromosome.BaseEncode(in_value);
        }

	else
	{
		double in_value;
		long num_genes = chromosome._num_genes;
		for (long g=0; g<num_genes; g++)
		{
			is >> in_value;
			chromosome.WriteGene(g, in_value);
		}
        }

	return chromosome;
}



/*****************************************************************
 Class Individual
 *****************************************************************/


Individual::Individual(void)
{
	_num_chromosomes = 0;
	_chromosome_array = 0;
	_fitness_valid = FALSE;
	_pool_index = INVALID_INDEX;
	_location._x = INVALID_INDEX;
	_location._y = INVALID_INDEX;
}



Individual::Individual(long num_chromosomes)
{
	_num_chromosomes = num_chromosomes;

	_chromosome_array = new Chromosome*[_num_chromosomes];
	for (long c=0; c<_num_chromosomes; c++)
		_chromosome_array[c] = 0;
	
	_fitness_valid = FALSE;
	_pool_index = INVALID_INDEX;
	_location._x = INVALID_INDEX;
	_location._y = INVALID_INDEX;
}



Individual::Individual(long num_chromosomes, long num_genes, GeneType type)
{
	_num_chromosomes = num_chromosomes;

	_chromosome_array = new Chromosome*[_num_chromosomes];
	for (long c=0; c<_num_chromosomes; c++)
		_chromosome_array[c] = new Chromosome(num_genes, type);
	
	_fitness_valid = FALSE;
	_pool_index = INVALID_INDEX;
	_location._x = INVALID_INDEX;
	_location._y = INVALID_INDEX;
}


Individual::Individual(long num_chromosomes, long num_genes, char alphabet)
{
	_num_chromosomes = num_chromosomes;

	_chromosome_array = new Chromosome*[_num_chromosomes];
	for (long c=0; c<_num_chromosomes; c++)
		_chromosome_array[c] = new Chromosome(num_genes, alphabet);
	
	_fitness_valid = FALSE;
	_pool_index = INVALID_INDEX;
	_location._x = INVALID_INDEX;
	_location._y = INVALID_INDEX;
}


Individual::Individual(const Individual& old_individual)
{
	_num_chromosomes = 0;
	_chromosome_array = 0;

	*this = old_individual;
}


Individual& Individual::operator=(const Individual& rhs)
{
	for (long c=0; c<_num_chromosomes; c++)
		delete _chromosome_array[c];
	delete[] _chromosome_array;

	_num_chromosomes = rhs._num_chromosomes;

	_chromosome_array = new Chromosome*[_num_chromosomes];
	for (c=0; c<_num_chromosomes; c++)
		_chromosome_array[c] = new Chromosome(*(rhs._chromosome_array[c]));

	_fitness_valid = rhs._fitness_valid;
	_fitness_cache = rhs._fitness_cache;

	_pool_index = rhs._pool_index;
	_location = rhs._location;

	return *this;
}


Individual::~Individual(void)
{
	for (long c=0; c<_num_chromosomes; c++)
		delete _chromosome_array[c];
	delete[] _chromosome_array;
}



/*****************************************************************
 Functions to allow writer of fitness function access to internal
 vars, given a particular individual object
 *****************************************************************/

long Individual::GetPoolIndex(void)
{       return _pool_index;	}


long Individual::GetNumChromosomes(void)
{       return _num_chromosomes;	}

long Individual::GetNumGenes(long chromosome)
{       return _chromosome_array[chromosome]->_num_genes;	}

GeneType Individual::GetGeneType(long chromosome)
{       return _chromosome_array[chromosome]->_type;	}

double Individual::GetMinValue(long chromosome)
{	return _chromosome_array[chromosome]->_min;	}

double Individual::GetMaxValue(long chromosome)
{	return _chromosome_array[chromosome]->_max;	}

/*************************************************************************
 NB: Invalidates pool order cache and fitness valid if it induces new value
 *************************************************************************/

void Individual::SetMinValue(long chromosome, double value)
{	_chromosome_array[chromosome]->_min = value; }

/*************************************************************************
 NB: Invalidates pool order cache and fitness valid if it induces new value
 *************************************************************************/

void Individual::SetMaxValue(long chromosome, double value)
{	_chromosome_array[chromosome]->_max = value;}



char Individual::GetAlphabet(long chromosome)
{	return _chromosome_array[chromosome]->_alphabet;	}



/*************************************************************************
 NB: Invalidates pool order cache
 *************************************************************************/

void Individual::AddChromosome(long index, Chromosome* chromosome)
{
	if (index < _num_chromosomes  && index >=0)
	{
		delete _chromosome_array[index];
		_chromosome_array[index] = chromosome;
	}
}

/*************************************************************************
 NB: Invalidates pool order cache
 *************************************************************************/

void Individual::SetNumChromosomes(long new_num)
{
	if (new_num != _num_chromosomes)
	{
		Chromosome** new_array = new Chromosome*[new_num];

		// copy old chromosomes
		int num_to_copy;
		if (new_num < _num_chromosomes)
			num_to_copy = new_num;
		else
			num_to_copy = _num_chromosomes;
		for (int c=0; c<num_to_copy; c++)
			new_array[c] = _chromosome_array[c];

                // delete any extra old ones
		if (new_num < _num_chromosomes)
                {
			for (c=num_to_copy; c<_num_chromosomes; c++)
				delete _chromosome_array[c];
		}
                // duplicate the last one if new array is bigger than old
		if (new_num > _num_chromosomes)
		{
			for (c=num_to_copy-1; c<new_num; c++)
			{
				Chromosome* new_chromosome = new Chromosome(*(_chromosome_array[num_to_copy-1]));
				new_array[c] = new_chromosome;
                	}
		}

		delete[] _chromosome_array;

		_chromosome_array = new_array;
		_num_chromosomes = new_num;
	}
}




double Individual::ReadChromosome(long chromosome)
{
	double value;

	if (chromosome <= _num_chromosomes)
        {  		
		if (_chromosome_array[chromosome]->_type == G_ENCODED)
				value = _chromosome_array[chromosome]->BaseDecode();
		else
			value = _chromosome_array[chromosome]->ReadGene(0);
	}

	return value;
}



/*************************************************************************
 NB: Invalidates pool order cache
 *************************************************************************/

void Individual::WriteChromosome(long chromosome, double value)
{
	if (chromosome <= _num_chromosomes)
        {  		
		if (_chromosome_array[chromosome]->_type == G_ENCODED)
			_chromosome_array[chromosome]->BaseEncode(value);
		else
			_chromosome_array[chromosome]->WriteGene(0, value);
		_fitness_valid = FALSE;
	}
}



double Individual::ReadGene(long chromosome, long gene)
{
	double value;

	if (chromosome <= _num_chromosomes
		&&  gene <= _chromosome_array[chromosome]->_num_genes)
	{
		switch (_chromosome_array[chromosome]->_type)
		{   
			case G_ENCODED:
				value = ((char*)(_chromosome_array[chromosome]->_gene_array))[gene];
				break;

			case G_INT:
				value = ((long*)(_chromosome_array[chromosome]->_gene_array))[gene];
				break;

			case G_FLOAT:
				value = ((double*)(_chromosome_array[chromosome]->_gene_array))[gene];
				break;

			case G_USER:
				value = ((Gene*)(_chromosome_array[chromosome]->_gene_array))[gene];
				break;
		}
	}
	return value;
}


FitnessType Individual::GetFitness(void)
{
	if (!_fitness_valid)
	{
		_fitness_cache = Algorithm::_evaluator->EvaluateFitness(*this);
		_fitness_valid = TRUE;
	}
	return _fitness_cache;
}



/*****************************************************************
 Location of individual (for sturctured pools)
 *****************************************************************/

Location Individual::GetLocation(void)
{	return _location;	}

void Individual::SetLocation(Location location)
{	_location = location;	}




/*****************************************************************
 iostreams
 *****************************************************************/

ostream& operator<< (ostream& os, const Individual& individual)
{
	for (long c=0; c<individual._num_chromosomes; c++)
		os << *(individual._chromosome_array[c]) << ' ';

	return os;
}


Individual& operator>> (istream& is, Individual& individual)
{
	long num_chromosomes = individual._num_chromosomes;
	for (long c=0; c<num_chromosomes; c++)
	{
		is >> *(individual._chromosome_array[c]);
	}

	return individual;
}



// outputs to the ostream an undecoded version of the individual

ostream& Individual::PrintGenotype(ostream& os)
{
	for (long c=0; c<_num_chromosomes; c++)
	{
		long num_genes = _chromosome_array[c]->_num_genes;
		for (long g=0; g<num_genes; g++)
                {
			os << ReadGene(c, g) << ' ';
		}
		os << "   ";
	}
	return os;
}


// outputs to the ostream a decoded version of the individual,
// including its fitness value

ostream& Individual::PrintPhenotype(ostream& os)
{
	long fitness = GetFitness();
	os << *this << '\t' << fitness;

	return os;
}
