// *******************************************************************************
// File name:		oplib.cpp
// Author:		Jose-Luiz Ribeiro-Filho, Laura Dekker
// Description:		Library of operators for GAs, hillclimbers, etc.
//
//
// Change history: WBL SUN gets % -ve numbers wrong
//
//
// *******************************************************************************
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <fstream.h>

#include "individ.h"
#include "pools.h"
#include "opids.h"
#include "oplib.h"
#include "rand.h"

#define ABS(x)  ((x)>=0? (x) : -(x))

#ifdef DEBUG1
extern ofstream debug;
#define debug cout
#endif

/**********************************************************************
  OPERATORS LIBRARY MANAGEMENT
 **********************************************************************/

void OperatorLibrary::CatalogueOperators(void)
{
	// put all selection operators here
	AddCatalogueEntry(OP_SELECTION, ROULETTE_WHEEL,	"Roulette Wheel");
	AddCatalogueEntry(OP_SELECTION, TRUNCATE_SELECT, "Truncated Selection");
	AddCatalogueEntry(OP_SELECTION, TOURNAMENT_SELECT, "Tournament Selection");
	AddCatalogueEntry(OP_SELECTION, LINEAR_SELECT, "Linear Selection");
	AddCatalogueEntry(OP_SELECTION, REMAIND_STOCHASTIC, "Remainder Stochastic");
	AddCatalogueEntry(OP_SELECTION, WINDOW_SELECT, "Window Select");

	// put all crossover operators here
	AddCatalogueEntry(OP_CROSSOVER, HOLLAND_CROSSOVER,"Holland Crossover");
	AddCatalogueEntry(OP_CROSSOVER, TWOPOINT_CROSSOVER, "Two Point Crossover");
	AddCatalogueEntry(OP_CROSSOVER, UNIFORM_CROSSOVER, "Uniform Crossover");
	AddCatalogueEntry(OP_CROSSOVER, COMBINATORIAL_CROSSOVER, "Combinatorial Crossover");

	// put all mutation operators here
	AddCatalogueEntry(OP_MUTATION, BASIC_MUTATION,"Basic Mutation");
	AddCatalogueEntry(OP_MUTATION, COMBINATORIAL_MUTATION,"Combinatorial Mutation");

	// put all selection operators here
	AddCatalogueEntry(OP_MIGRATION, BEST_TO_WORST_MIGRATION, "Best-To-Worst Migration");

	// put all initialisation operators here
	AddCatalogueEntry(OP_INITIALISE, RANDOM_INITIALISE, "Random Initialisation");
	AddCatalogueEntry(OP_INITIALISE, INIT_FROM_FILE, "Initialise From File");
	AddCatalogueEntry(OP_INITIALISE, INIT_COMBINATORIAL, "Combinatorial Initialisation");

	// put all hill-climbing operators here
	AddCatalogueEntry(OP_HILL_CLIMBER, HILL_CLIMB3, "Hill Climber #3");

	// put all other operators here

}



///////////////////////////////////////////////////////////////////////
// Creates an instance of an operator with the given ID, and returns 
// the index into the instatiated operators array
//
// Returns INVALID_INDEX if no match.
//

int OperatorLibrary::CreateOperator(OperatorID operatorID)
{
	BaseOperator* new_operator = 0;

	switch (operatorID)
	{
        	// Initialisation operators
		case RANDOM_INITIALISE:			
			new_operator = new RandomInitialise; break;
		case INIT_FROM_FILE:
			new_operator = new InitFromFile; break;
		case INIT_COMBINATORIAL:
			new_operator = new CombinatorialInitialise; break;

		// Selection Operators
		case TRUNCATE_SELECT:
			new_operator = new TruncatedSelection; break;
		case ROULETTE_WHEEL:
			new_operator = new RouletteWheel; break;
		case TOURNAMENT_SELECT:
			new_operator = new TournamentSelection; break;
		case LINEAR_SELECT:
			new_operator = new LinearSelection; break;
		case REMAIND_STOCHASTIC:
			new_operator = new RemainderStochastic; break;
		case WINDOW_SELECT:
			new_operator = new WindowSelection; break;

		// Crossover operators
		case HOLLAND_CROSSOVER:
			new_operator = new HollandCrossover; break;
		case TWOPOINT_CROSSOVER:
			new_operator = new TwoPointCrossover; break;
		case UNIFORM_CROSSOVER:
			new_operator = new UniformCrossover; break;
		case COMBINATORIAL_CROSSOVER:
			new_operator = new CombinatorialCrossover; break;

		// Mutation operators				
		case BASIC_MUTATION:
			new_operator = new BasicMutation; break;
		case COMBINATORIAL_MUTATION:
			new_operator = new CombinatorialMutation; break;

		// Migration operators
		case BEST_TO_WORST_MIGRATION:
			new_operator = new BestToWorstMigration; break;

		default:
			new_operator = 0; break;
	}

	if (new_operator)
	{
		// get the operator name from the catalogue
		OperatorType type = new_operator->GetOperatorType();
		int index = GetOperatorIndex(type, operatorID);		
		new_operator->SetOperatorName(GetOperatorName(type, index));

		// find the first gap and add the new operator there
		for (int i=0; i<MAX_OPERATORS; i++)
		{
			if (!_instantiated_operators[i])
                        {
				_instantiated_operators[i] = new_operator;
				break;
			}
		}
                return i;
	}
	else	
		return INVALID_INDEX;
}






/********************************************************************
 Operator Library
 ********************************************************************/

OperatorLibrary::OperatorLibrary(void)
{
	_num_catalogued = 0;	

	_num_initialisation = 0;
	_num_selection = 0;
	_num_crossover = 0;
	_num_mutation = 0;
	_num_migration = 0;
	_num_hillclimber = 0;
	_num_misc = 0;

	for (int i=0; i<MAX_OPERATORS; i++)
		_instantiated_operators[i] = 0;

	CatalogueOperators();
}


OperatorLibrary::~OperatorLibrary(void)
{
	for (int i=0; i<_num_catalogued; i++)
		delete _catalogue[i];
	for (i=0; i<MAX_OPERATORS; i++)
		delete _instantiated_operators[i];
}


void OperatorLibrary::DeleteOperator(int operator_index)
{
	if (operator_index < MAX_OPERATORS  &&  operator_index >= 0)
        {
		delete	_instantiated_operators[operator_index];
		_instantiated_operators[operator_index] = 0;
	}
}


void OperatorLibrary::AddCatalogueEntry(OperatorType type, OperatorID ID, char* name)
{
	OperatorLibraryEntry* new_entry = new OperatorLibraryEntry;

	new_entry->_type = type;
	new_entry->_ID = ID;
	strncpy(new_entry->_name, name, LABEL_LENGTH-1);
	new_entry->_name[LABEL_LENGTH-1] = 0;

	_catalogue[_num_catalogued] = new_entry;
	
	switch (type)
	{
		case OP_INITIALISE:
			_num_initialisation++; break;
		case OP_SELECTION:
			_num_selection++; break;
		case OP_CROSSOVER:
			_num_crossover++; break;
		case OP_MUTATION:
			_num_mutation++; break;
		case OP_MIGRATION:
			_num_migration++; break;
		case OP_HILL_CLIMBER:
			_num_hillclimber++; break;
		default:
			_num_misc++; break;
	}

	_num_catalogued++;
}




BaseOperator* OperatorLibrary::GetOperator(int operator_index)
{
	if (operator_index<MAX_OPERATORS  &&  operator_index>=0)
		return	_instantiated_operators[operator_index];
	else
        	return 0;
}



int OperatorLibrary::GetNumOperators(OperatorType type)
{
	switch (type)
	{
		case OP_INITIALISE:
			return _num_initialisation;
		case OP_SELECTION:
			return _num_selection;
		case OP_CROSSOVER:
			return _num_crossover;
		case OP_MUTATION:
			return _num_mutation;
		case OP_MIGRATION:
			return _num_migration;
		case OP_HILL_CLIMBER:
			return _num_hillclimber;
		default:
			return _num_misc;
	}
}



char* OperatorLibrary::GetOperatorName(OperatorType type, int index)
{
	index = GetStartIndex(type)+index;
	return _catalogue[index]->_name;
}



OperatorID OperatorLibrary::GetOperatorID(OperatorType type, int index)
{
	index = GetStartIndex(type)+index;
	return _catalogue[index]->_ID;
}



int OperatorLibrary::GetOperatorIndex(OperatorType type, OperatorID ID)
{
	int start_index = GetStartIndex(type);
	int end_index = start_index+GetNumOperators(type);

	for (int i=start_index; i<end_index; i++)
	{
		if (_catalogue[i]->_ID == ID)
			break;
	}

	if (i>=end_index) // if we've got to the end of the array without finding a match...
		return INVALID_INDEX;
	else		// otherwise return the index where the match was found
		return i-start_index;
}



int OperatorLibrary::GetOperatorIndex(OperatorType type, char* name)
{
	int start_index = GetStartIndex(type);
	int end_index = start_index+GetNumOperators(type);

	for (int i=start_index; i<end_index; i++)
	{
		if (!strcmp(_catalogue[i]->_name, name))
			break;
	}

	if (i>=end_index) // if we've got to the end of the array without finding a match...
		return INVALID_INDEX;
	else		// otherwise return the index where the match was found
		return i-start_index;
}



int OperatorLibrary::GetStartIndex(OperatorType type)
{
	int start_index;

	switch (type)
        {
		case OP_INITIALISE:
			start_index = _num_migration+_num_mutation+_num_crossover+
					_num_selection; break;
		case OP_SELECTION:
			start_index = 0; break;
		case OP_CROSSOVER:
			start_index = _num_selection; break;
		case OP_MUTATION:
			start_index = _num_crossover+_num_selection; break;
		case OP_MIGRATION:
			start_index = _num_mutation+_num_crossover+_num_selection; break;
		case OP_HILL_CLIMBER:
			start_index = _num_initialisation+_num_migration+_num_mutation+
					_num_crossover+_num_selection; break;
		case OP_MISC:
			start_index = _num_hillclimber+_num_initialisation+_num_migration+
					_num_mutation+_num_crossover+_num_selection; break;
	}
	return start_index;
}








/**************************************************************************
 Base Operator class
 **************************************************************************/

BaseOperator::~BaseOperator(void)
{
	for (int i=0; i<_num_parameters; i++)
	{
		if (_parameter_array[i]._type == P_FILENAME)
                {
			char* string = (char*)((long)(_parameter_array[i]._value));
			delete[] string;
		}
        }
}


BaseOperator::BaseOperator(void)
{
	_name[0] = 0;
	_type = OP_MISC;	// default type;
	_num_parameters = 0;
	_ID = 0;
} 

void BaseOperator::SetOperatorName(char* name)
{
	strncpy(_name, name, LABEL_LENGTH-1);
	_name[LABEL_LENGTH-1] = 0;
} 

char* BaseOperator::GetOperatorName(void)
{	return _name;	} 

OperatorID BaseOperator::GetOperatorID(void)
{	return _ID;	}



/******************************************************************
 Get and set the operator type.  The operator type is used to
 classify it, eg. in UI displays for selecting operators
 ******************************************************************/

void BaseOperator::SetOperatorType(OperatorType type)
{	_type = type;	}     

OperatorType BaseOperator::GetOperatorType(void)
{	return _type;	}


       
/******************************************************************
 Add a user-settable parameter to the operator; this information is
 used by the GUI to generate automatic dialog boxes for parameter
 setting, etc.
 Description and type are used by the dialog box.
 This is only necessary for user-settable parameters - parameters
 set by the algorithm itself can be done via normal variables
 ******************************************************************/

int BaseOperator::AddNumericParameter(char* description, ParameterType type, double default_value)
{
	if (_num_parameters >= MAX_PARAMS)
		return INVALID_INDEX;
        
	strncpy(_parameter_array[_num_parameters]._description, description, LABEL_LENGTH);
	_parameter_array[_num_parameters]._type = type;
	_parameter_array[_num_parameters]._value = default_value;

	// return the index to this parameter
	return _num_parameters++;
}


int BaseOperator::AddStringParameter(char* description, ParameterType type, char* default_value)
{
	if (_num_parameters >= MAX_PARAMS)
		return INVALID_INDEX;
        
	strncpy(_parameter_array[_num_parameters]._description, description, LABEL_LENGTH);
	_parameter_array[_num_parameters]._type = type;
	char* string_ptr = new char[MAX_FILE_PATH];
	strcpy(string_ptr, default_value);
	_parameter_array[_num_parameters]._value = ((double)((long)string_ptr));

	// return the index to this parameter
	return _num_parameters++;
}


char BaseOperator::GetNumParameters(void)
{
	return _num_parameters;
}


char* BaseOperator::GetParameterDescription(int index)
{
	if (index < 0  ||  index > _num_parameters)
		return 0;

	return _parameter_array[index]._description;
}


ParameterType BaseOperator::GetParameterType(int index)
{
	assert(index >= 0  &&  index <= _num_parameters);
	return _parameter_array[index]._type;
}



/******************************************************************
 Always returns a double - caller does any type conversion necessary
 ******************************************************************/

double BaseOperator::GetParameterValue(int index)
{
	if (index < 0  ||  index > _num_parameters)
		return 0;

	return _parameter_array[index]._value;
}


void BaseOperator::SetParameterValue(int index, double value)
{
	if (index < 0  ||  index > _num_parameters)
		return;

	_parameter_array[index]._value = value;
}

char* BaseOperator::GetParameterString(int index)
{
	if (index < 0  ||  index > _num_parameters)
		return 0;

	return (char*)((long)(_parameter_array[index]._value));
}


void BaseOperator::SetParameterString(int index, char* value)
{
	if (index < 0  ||  index > _num_parameters)
		return;

	char* current_string = (char*)((long)(_parameter_array[index]._value));
	strcpy(current_string, value);
}







/**************************************************************************
 Base Crossover class
 **************************************************************************/

Crossover::Crossover(void)
{
	SetOperatorName("Base Crossover");
	_type = OP_CROSSOVER;
	_rate_param = AddNumericParameter("Crossover rate", P_FLOAT, DEF_PCROSS);
}


/**************************************************************************
 Base Mutation class
 **************************************************************************/

Mutation::Mutation(void)
{
	SetOperatorName("Base Mutation");
	_type = OP_MUTATION;
	_rate_param = AddNumericParameter("Mutation rate", P_FLOAT, DEF_PMUT);
}


/**************************************************************************
 Base Migration class
 **************************************************************************/

Migration::Migration(void)
{
	SetOperatorName("Base Migration");
	_type = OP_MIGRATION;
	_rate_param = AddNumericParameter("Migration rate", P_FLOAT, DEF_PMIG);
}


/**************************************************************************
 Base Selection class
 **************************************************************************/

Selection::Selection(void)
{
	SetOperatorName("Base Selection");
	_type = OP_SELECTION;
}


/**************************************************************************
 Base Initialisation class
 **************************************************************************/

Initialisation::Initialisation(void)
{
	SetOperatorName("Base Initialisation");
	_type = OP_INITIALISE;
}

 
/**************************************************************************
 Base HillClimbing class
 **************************************************************************/

HillClimbing::HillClimbing(void)
{
	SetOperatorName("Base HillClimbing");
	_type = OP_HILL_CLIMBER;
}





/**********************************************************************
 INITIALISATION OPERATORS
 **********************************************************************/

///////////////////////////////////////////////////////////////////////
// Random Initialisation
//

RandomInitialise::RandomInitialise(void)
{ 	
	_ID = RANDOM_INITIALISE;
}

void RandomInitialise::DoInitialisation(Pool* pool)
{
	long num_individuals = pool->GetNumIndividuals();
	for (long i=0; i<num_individuals; i++)
	{
		long num_chromosomes = pool->GetNumChromosomes(i);
		for (long c=0; c<num_chromosomes; c++)
		{
			double min = pool->GetMinValue(i, c);
			double max = pool->GetMaxValue(i, c);

			if (pool->GetGeneType(i, c) == G_ENCODED)
			{
				pool->GetChromosomeHandle(i, c)->BaseEncode(random((unsigned long)(max+1)));
			}
			else
                        {
				long num_genes = pool->GetNumGenes(i, c);
				for (long g=0; g<num_genes; g++)
					pool->WriteGene(i,c,g, min+random((unsigned long)((max+1)-min)));
			}
		}
	} 	
}



///////////////////////////////////////////////////////////////////////
// Initialise from a file


InitFromFile::InitFromFile(void)
{
	_ID = INIT_FROM_FILE;
	_filename_param = AddStringParameter("File name", P_FILENAME, "");
}


void InitFromFile::DoInitialisation(Pool* pool)
{
// Note: Currently only uses the values for the first pool supplied,
//       so that multiple pools will start with duplicate values

	ifstream input_file(GetParameterString(_filename_param));

	// if at end of file wrap around to start again
	if (input_file.eof())
		input_file.seekg(0);
	input_file >> *pool;
}




///////////////////////////////////////////////////////////////////////
// Initialisation for order-based algorithms

CombinatorialInitialise::CombinatorialInitialise(void)
{
	_ID = INIT_COMBINATORIAL;
}

void CombinatorialInitialise::DoInitialisation(Pool* pool)
{
	int flag;
	long new_value;
	
	long num_individuals = pool->GetNumIndividuals();
	for (int i=0; i<num_individuals; i++)
	{
		long num_chromosomes = pool->GetNumChromosomes(i);
		for (long c=0; c<num_chromosomes; c++)
		{
			long num_genes = pool->GetNumGenes(i, c);
			for (long g=0; g<num_genes; g++)
			{
				do
				{
					flag = 0;
					new_value = random(num_genes);

					for (long j=0; j<g; j++)
                                	{
						if (new_value == pool->ReadGene(i, c, j))
						{
							flag = 1;
							continue;
						}
					}				
				}
				while (flag == 1);
				pool->WriteGene(i, c, g, new_value);
			}
		}
	}
}





/**********************************************************************
 SELECTION OPERATORS
 **********************************************************************/

///////////////////////////////////////////////////////////////////////
// Truncated Selection
//

TruncatedSelection::TruncatedSelection(void)
{	
	_ID = TRUNCATE_SELECT;
	_truncate_param = AddNumericParameter("Truncation", P_FLOAT, DEF_TRUNCATION);
	_norm_scale_param = AddNumericParameter("Normalising Scale", P_LONG, DEF_NORMALISING_SCALE);
}

void TruncatedSelection::DoSelection(Pool* pool, OptimiseType min_or_max)
{     
	const long NORMALISING_SCALE = GetParameterValue(_norm_scale_param);

	// assumes mating pool size == pool size
#ifdef DEBUG1
	debug << "Selecting mating pool\n";
#endif
	long src_i, dst_i;
	long roulette, parsum;
	FitnessType total_fitness;

	long num_individuals = pool->GetNumIndividuals();

	// create the mating pool
	Pool mating_pool(num_individuals);

	// determine number of top individuals automatically selected
	long num_auto_select = num_individuals*GetParameterValue(_truncate_param);
	long* order = new long[num_auto_select];

	// copy the indices of the best to a temporary array
	for (long i=0; i<num_auto_select; i++)
		order[i] = pool->GetNthHighest(i, min_or_max);

	// copy the best to mating pool in random order
	for (dst_i=0; dst_i<num_auto_select; dst_i++)
	{
		src_i=random(num_auto_select);
		while (order[src_i]==EMPTY)
		{
			if(++src_i >= num_auto_select)
				src_i=0;
		}
		CopyIndividual(pool, order[src_i], &mating_pool, dst_i);
		order[src_i]=EMPTY;
	}

	// now do the rest by roulette
	total_fitness = pool->GetTotalFitness();

	if (min_or_max == MINIMISE)	// if minimising
	{
		FitnessType fitness;
		FitnessType min = pool->GetFitness(pool->GetHighestFitness(min_or_max));
		FitnessType max = pool->GetFitness(pool->GetLowestFitness(min_or_max));
		FitnessType diff = max-min;

		// Normalise and invert
		for (long double norm_total=0, i=0; i<num_individuals; i++)
		{
			fitness = pool->GetFitness(i);
			if (fitness == min)
				norm_total += 1.0;
			else
				norm_total += 1.0-(fitness-min)/diff;
		}

		for (dst_i=num_auto_select; dst_i<num_individuals; dst_i++)
		{
			long rand1 = rand();
			long rand2 = rand();
			roulette = (ABS(rand1*rand2)) % ((long)(norm_total*NORMALISING_SCALE)); // spin the roulette wheel

			for (parsum=0, src_i=0; (src_i<num_individuals) && (parsum<=roulette); src_i++)
			{
				fitness = pool->GetFitness(src_i);
				if (fitness == min)
					fitness = NORMALISING_SCALE;
				else				
					fitness = NORMALISING_SCALE*(1.0-((fitness-min)/diff));

				parsum += (long)fitness;
			}

			CopyIndividual(pool, --src_i, &mating_pool, dst_i);
		}
	}
	else	// maximising
	{
	#ifdef DEBUG1
		debug << "total fitness = " << total_fitness << endl;
	#endif
		for (dst_i=num_auto_select; dst_i<num_individuals; dst_i++)
		{
			long rand1 = rand();
			long rand2 = rand();
			long randnum = ABS(rand1*rand2);	
			roulette = randnum % ((long)total_fitness);   // spin the roulette wheel			
	#ifdef DEBUG1
		debug << "rand1, rand2, roulette = " 
		      << rand1<<", "<<rand2<<", "<<roulette << endl;
	#endif
                
			for (parsum=0, src_i=0; (src_i<num_individuals) && (parsum<=roulette); src_i++)
				parsum += pool->GetFitness(src_i);
			
			CopyIndividual(pool, --src_i, &mating_pool, dst_i);
		}
	}

	// now copy the new individual array back to the pool
	for (i=0; i<num_individuals; i++)
		pool->AddIndividual(i, mating_pool.RemoveIndividual(i));

	delete[] order; 
}

///////////////////////////////////////////////////////////////////////
// Truncated Selection2
//
// This is just a copy of the above class <<<< Must be modified!
//

TruncatedSelection2::TruncatedSelection2(void)
{ 	
	_ID = TRUNCATE_SELECT2;
	_truncate_param = AddNumericParameter("Truncation", P_FLOAT, DEF_TRUNCATION);
}

void TruncatedSelection2::DoSelection(Pool* pool, OptimiseType min_or_max)
{
	// TODO: insert the code here
}



///////////////////////////////////////////////////////////////////////
// Roulette Wheel
//

RouletteWheel::RouletteWheel(void)
{
	SetOperatorName("Roulette Wheel");
	_ID = ROULETTE_WHEEL;
	_norm_scale_param = AddNumericParameter("Normalising Scale", P_LONG, DEF_NORMALISING_SCALE);
	_elitism_flag_param = AddNumericParameter("Elitism", P_BOOL, DEF_ELITISM);
}

void RouletteWheel::DoSelection(Pool* pool, OptimiseType min_or_max)
{     
	const long NORMALISING_SCALE = GetParameterValue(_norm_scale_param);;
	BOOL elitism = GetParameterValue(_elitism_flag_param);

#ifdef DEBUG1
	debug << "Selecting mating pool\n";
#endif

	long num_individuals = pool->GetNumIndividuals();
	FitnessType min = pool->GetFitness(pool->GetHighestFitness(min_or_max));
	FitnessType max = pool->GetFitness(pool->GetLowestFitness(min_or_max));
	FitnessType fitness;
	FitnessType total_fitness = 0;

    // Get the total fitness (for individuals scaled between 1.0 and 2.0)
	for (long i=0; i<num_individuals; i++)
    {
		fitness = MinOrMax(min_or_max,pool->GetFitness(i),min,max);
		total_fitness += fitness;
    }

	long src_i, dst_i;
	long roulette, parsum;

	// create the mating pool
	// assumes mating pool size == pool size
	Pool mating_pool(num_individuals);

	for (i=0; i<num_individuals; i++)
	{
		long rand1 = rand();
		long rand2 = rand();
		long randnum = ABS(rand1*rand2);

		long roulette = randnum % (long)(NORMALISING_SCALE*total_fitness);   // spin the roulette wheel
                
		for (long parsum=0, src_i=0; (src_i<num_individuals) && (parsum<=roulette); src_i++)
		{
			fitness = MinOrMax(min_or_max,pool->GetFitness(i),min,max);
			parsum += (long)(NORMALISING_SCALE*fitness);
		}
		CopyIndividual(pool,--src_i,&mating_pool,i);
	}

	// if elitism is used then replace the first selected individual
	// by the best individual 
	if (elitism)
	{
		src_i= pool->GetNthHighest(0, min_or_max);
		mating_pool.RemoveIndividual(0);
		CopyIndividual(pool,src_i,&mating_pool,0);
	}

	// now copy the new individual array back to the pool
	for (i=0; i<num_individuals; i++)
		pool->AddIndividual(i, mating_pool.RemoveIndividual(i));
}

///////////////////////////////////////////////////////////////////////
// Tournament Selection
//

TournamentSelection::TournamentSelection(void)
{
	_ID = TOURNAMENT_SELECT;
	_grp_sz_param = AddNumericParameter("Group Size", P_LONG, DEF_GROUP_SIZE);
}

void TournamentSelection::DoSelection(Pool* pool, OptimiseType min_or_max)
{     
	const long GROUP_SIZE = GetParameterValue(_grp_sz_param);

#ifdef DEBUG1
	debug << "Selecting mating pool\n";
#endif

	long num_individuals = pool->GetNumIndividuals();
	FitnessType min = pool->GetFitness(pool->GetHighestFitness(min_or_max));
	FitnessType max = pool->GetFitness(pool->GetLowestFitness(min_or_max));
	FitnessType fitness;

	// create the mating pool
	// assumes mating pool size == pool size
	Pool mating_pool(num_individuals);

	// For the whole population
	for (long next=0; next<num_individuals; next++)
    {
		// Randomly pick a group of individuals up from the population
		FitnessType fitness, best_fitness;
		long index, selected;

        // get the first member of the group and start the search
		long rand1 = rand();
		long rand2 = rand();
		long randnum = ABS(rand1*rand2);

		selected = randnum % ((long)num_individuals);
		best_fitness = MinOrMax(min_or_max,pool->GetFitness(selected),min,max);

        for (long i=1; i<GROUP_SIZE; i++)
		{
			rand1 = rand();
			rand2 = rand();
			randnum = ABS(rand1*rand2);

			index = randnum % ((long)num_individuals);
			fitness = MinOrMax(min_or_max,pool->GetFitness(index),min,max);

			// Keep the index of the best member
			if (fitness > best_fitness)
			{
				best_fitness = fitness;
				selected = index;
			}
		}

		// Put the best of this group in the mating pool
 		CopyIndividual(pool,selected,&mating_pool,next);
	}
	// now copy the new individual array back to the pool
	for (long i=0; i<num_individuals; i++)
		pool->AddIndividual(i,mating_pool.RemoveIndividual(i));
}

///////////////////////////////////////////////////////////////////////
// Linear Selection
//

LinearSelection::LinearSelection(void)
{ 	
	_ID = LINEAR_SELECT;
	_constant_param = AddNumericParameter("Constant", P_INT, DEF_LIN_CONST);
	_gradient_param = AddNumericParameter("Gradient", P_INT, 0);
}

void LinearSelection::DoSelection(Pool* pool, OptimiseType min_or_max)
{
	double initial_constant = GetParameterValue(_constant_param);
	double gradient = GetParameterValue(_gradient_param);

	long num_individuals = pool->GetNumIndividuals();

	// create the mating pool
	Pool mating_pool(num_individuals);

	if (gradient == 0)	// Not defined !
		gradient = initial_constant/num_individuals;

	// Make the descending ordered array
	double total_fitness = initial_constant;
	double* padFitness = new double[num_individuals];
	padFitness[0] = initial_constant;

	for (long i=1; i<num_individuals; i++)
    {
		padFitness[i] = padFitness[i-1]-gradient;
		total_fitness += padFitness[i];
    }

	// Apply roullete
	long dst_i = 0;
	for (i = 0; i<num_individuals; i++)
	{
 		long rand1 = rand();
		long rand2 = rand();
		long randnum = ABS(rand1*rand2);

		long roulette = randnum % ((long)total_fitness);   // spin the roulette wheel
                
		for (long parsum=0, src_i=0; (src_i<num_individuals) && (parsum<=roulette); src_i++)
			parsum += (long)padFitness[src_i];
	
		CopyIndividual(pool,--src_i,&mating_pool,dst_i++);
	}

	// now copy the mating pool back to the original pool
	for (i=0; i<num_individuals; i++)
		pool->AddIndividual(i, mating_pool.RemoveIndividual(i));


	delete [] padFitness;
}

///////////////////////////////////////////////////////////////////////
// Remainder Stochastic
//

RemainderStochastic::RemainderStochastic(void)
{   	
	_ID = REMAIND_STOCHASTIC;
}

void RemainderStochastic::DoSelection(Pool* pool, OptimiseType min_or_max)
{
#ifdef DEBUG1
	debug << "Selecting mating pool"<< endl;
#endif

	long num_individuals = pool->GetNumIndividuals();
	FitnessType min = pool->GetFitness(pool->GetHighestFitness(min_or_max));
	FitnessType max = pool->GetFitness(pool->GetLowestFitness(min_or_max));
	FitnessType fitness;
	FitnessType total_fitness = 0;

	// Get the total fitness (for individuals scaled between 1.0 and 2.0)
	for (long i=0; i<num_individuals; i++)
    {
		fitness = MinOrMax(min_or_max,pool->GetFitness(i),min,max);
		total_fitness += fitness;
    }

	long occurrences = 0;
    
	// create the mating pool
	// assumes mating pool size == pool size
	Pool mating_pool(num_individuals);
	long mating_pool_pop   = 0;

	// Copy to the next generation according to each member's occurrence
    // expectation
	for (i=0; i<num_individuals; i++)
	{
		fitness = MinOrMax(min_or_max,pool->GetFitness(i),min,max);
		occurrences = (unsigned long)((fitness/total_fitness)*num_individuals);
		for (long next=0; next<occurrences; next++)
			CopyIndividual(pool, i, &mating_pool, next+mating_pool_pop);

		mating_pool_pop += occurrences;
	}

	// Top-up with roulette
	for (i = mating_pool_pop; i<num_individuals; i++)
	{
		long rand1 = rand();
		long rand2 = rand();
		long randnum = ABS(rand1*rand2);

		long roulette = randnum % (long)(DEF_NORMALISING_SCALE*total_fitness);   // spin the roulette wheel
                
		for (long parsum=0, src_i=0; (src_i<num_individuals) && (parsum<=roulette); src_i++)
		{
			fitness = MinOrMax(min_or_max,pool->GetFitness(i),min,max);
			parsum += (long)(DEF_NORMALISING_SCALE*fitness);
		}
		CopyIndividual(pool,--src_i,&mating_pool,i);
	}

	// now copy the mating pool back to the original pool
	for (i=0; i<num_individuals; i++)
		pool->AddIndividual(i, mating_pool.RemoveIndividual(i));

}
///////////////////////////////////////////////////////////////////////
// Window Selection
//

WindowSelection::WindowSelection(void)
{  	
	_ID = WINDOW_SELECT;
}

void WindowSelection::DoSelection(Pool* pool, OptimiseType min_or_max)
{
#ifdef DEBUG1
	debug << "Selecting mating pool"<< endl;
#endif

	long num_individuals = pool->GetNumIndividuals();
	FitnessType min = pool->GetFitness(pool->GetHighestFitness(min_or_max));
	FitnessType max = pool->GetFitness(pool->GetLowestFitness(min_or_max));

	// create the mating pool
	// assumes mating pool size == pool size
	Pool mating_pool(num_individuals);

	FitnessType* padFitness = new FitnessType[num_individuals];
	FitnessType min_fitness = 0.0;	// normalised values
	FitnessType total_fitness = 0;

	// Get the fitness (for individuals nomalised between 1.0 and 2.0)
	for (long i=0; i<num_individuals; i++)
    {
		padFitness[i] = MinOrMax(min_or_max,pool->GetFitness(i),min,max) - min_fitness;
		total_fitness += padFitness[i];
    }

	// Now select with roulette
	for (i=0; i<num_individuals; i++)
	{
		long rand1 = rand();
		long rand2 = rand();
		long randnum = ABS(rand1*rand2);

		long roulette = randnum % (long)(DEF_NORMALISING_SCALE*total_fitness);   // spin the roulette wheel
                
		for (long parsum=0, src_i=0; (src_i<num_individuals) && (parsum<=roulette); src_i++)
		{
			parsum += (long)(DEF_NORMALISING_SCALE * padFitness[i]);
		}
		CopyIndividual(pool,--src_i,&mating_pool,i);
	}

	// now copy the mating pool back to the original pool
	for (i=0; i<num_individuals; i++)
		pool->AddIndividual(i, mating_pool.RemoveIndividual(i));

	delete [] padFitness;
}

/**********************************************************************
 CROSSOVER OPERATORS
 **********************************************************************/

///////////////////////////////////////////////////////////////////////
// Holland Crossover = One Point Crossover
//

HollandCrossover::HollandCrossover(void)
{	
	_ID = HOLLAND_CROSSOVER;
}

void HollandCrossover::DoCrossover(Pool* pool)
{
#ifdef DEBUG1
	debug << "Doing crossover\n";
#endif
	long site;

	long num_individuals = pool->GetNumIndividuals();
	for (long i=0; i<num_individuals-1; i+=2)
	{
		long num_chromosomes = pool->GetNumChromosomes(i);
		for (long c=0; c<num_chromosomes; c++)
		{
			long num_genes = pool->GetNumGenes(i, c);
			if (Flip(GetParameterValue(_rate_param)))
			{			
				site = random(num_genes-1)+1;
				SwapGenes(pool,i,c,pool,i+1,c,site);	
			}
		}
	}
}

///////////////////////////////////////////////////////////////////////
// Two-Point Crossover
//

TwoPointCrossover::TwoPointCrossover(void)
{   	
	_ID = TWOPOINT_CROSSOVER;
}

void TwoPointCrossover::DoCrossover(Pool* pool)
{
#ifdef DEBUG1
	debug << "Doing crossover\n";
#endif
	long site;
	long length;

	long num_individuals = pool->GetNumIndividuals();
	for (long i=0; i<num_individuals-1; i+=2)
	{
		long num_chromosomes = pool->GetNumChromosomes(i);
		for (long c=0; c<num_chromosomes; c++)
		{
			long num_genes = pool->GetNumGenes(i, c);
			if (Flip(GetParameterValue(_rate_param)))
			{			
				site = random(num_genes-1)+1;
				length = random(num_genes-site)+1;
				SwapGenes(pool,i,c,pool,i+1,c,site,length);
			}
		}
	}
}

///////////////////////////////////////////////////////////////////////
// Uniform Crossover
//

UniformCrossover::UniformCrossover(void)
{ 	
	_ID = UNIFORM_CROSSOVER;
}

void UniformCrossover::DoCrossover(Pool* pool)
{
#ifdef DEBUG1
	debug << "Doing crossover\n";
#endif

	long num_individuals = pool->GetNumIndividuals();
	for (long i=0; i<num_individuals-1; i+=2)
	{
		long num_chromosomes = pool->GetNumChromosomes(i);
		for (long c=0; c<num_chromosomes; c++)
		{
			if (Flip(GetParameterValue(_rate_param)))
			{			
 				long num_genes = pool->GetNumGenes(i, c);

				for (long g=0; g<num_genes; g++)
				{
					// Only swap if tails
					if (!Flip(0.5))
						SwapGenes(pool,i,c,pool,i+1,c,g,1);
                }
			}
		}
	}
}



///////////////////////////////////////////////////////////////////////
// Combinatorial Crossover
//

CombinatorialCrossover::CombinatorialCrossover(void)
{ 	
	_ID = COMBINATORIAL_CROSSOVER;
}

void CombinatorialCrossover::DoCrossover(Pool* pool)
{
#ifdef DEBUG1
	debug << "Doing combinatorial crossover\n";
#endif

	long num_individuals = pool->GetNumIndividuals();
	for (long i=0; i<num_individuals-1; i+=2)
	{
		long num_chromosomes = pool->GetNumChromosomes(i);
		for (long c=0; c<num_chromosomes; c++)
		{
			if (Flip(GetParameterValue(_rate_param)))
			{
				// copy the original chromosomes
				Chromosome* old_chromosome1 = pool->GetChromosomeCopy(i, c);
				Chromosome* old_chromosome2 = pool->GetChromosomeCopy(i+1, c);

				long num_genes = pool->GetNumGenes(i, c);
				long site = random(num_genes-1)+1;

				SwapGenes(pool, i, c, pool, i+1, c, site); // perform one point crossover

				// Check for duplications
				for (long k=0; k<site; k++)	// scan crossover area for individual1
				{
					long j = site;
					while (j<num_genes)
					{ 						
						if (pool->ReadGene(i, c, k) == pool->ReadGene(i, c, j)) // duplicated
						{   							
							pool->WriteGene(i, c, k, old_chromosome1->ReadGene(j)); // replace it
							j = site;
						}
						else
							j++;
					}
				}

				for (k=0; k<site; k++)	// scan crossover area area for individual2
				{
					long j = site;
					while (j<num_genes)
					{ 						
						if (pool->ReadGene(i+1, c, k) == pool->ReadGene(i+1, c, j)) // duplicated
						{							
							pool->WriteGene(i+1, c, k, old_chromosome2->ReadGene(j)); // replace it
							j = site;
						}
						else
							j++;
					}
				}

				delete old_chromosome1;
				delete old_chromosome2;
			} 
		}
	}
}



/**********************************************************************
 MUTATION OPERATORS
 **********************************************************************/

///////////////////////////////////////////////////////////////////////
// Basic Mutation
//

BasicMutation::BasicMutation(void)
{  	
	_ID = BASIC_MUTATION;
}   

void BasicMutation::DoMutation(Pool* pool)
{   
#ifdef DEBUG1
	debug << "Doing mutation\n";
#endif
	int value, change;

	long num_individuals = pool->GetNumIndividuals();
	for (int i=0; i<num_individuals; i++)
	{
		long num_chromosomes = pool->GetNumChromosomes(i);
		for (int c=0; c<num_chromosomes; c++)
		{
			long num_genes = pool->GetNumGenes(i, c);
			for (int g=0; g<num_genes; g++)
			{
				if (Flip(GetParameterValue(_rate_param)))
				{
					int new_value;
									
					// if encoded, get random value between 0,alpha-1 
					if (pool->GetGeneType(i,c) == G_ENCODED)
					{
						new_value = Int_Rand(0, pool->GetAlphabet(i,c)-1);
					}

					// if not encoded, get random value between min,max
					else
					{
						int min = pool->GetMinValue(i,c);
						int max = pool->GetMaxValue(i,c);
						new_value = Int_Rand(min,max);
					}

					pool->WriteGene(i,c,g, new_value);
					#ifdef DEBUG2
					debug << "Writing new value " << new_value << " at pos " << g << endl;
					#endif
				}
			}
		}
	}
}




///////////////////////////////////////////////////////////////////////
// Combinatorial Mutation
//
CombinatorialMutation::CombinatorialMutation(void)
{  	
	_ID = COMBINATORIAL_MUTATION;
}


void CombinatorialMutation::DoMutation(Pool* pool)
{   
#ifdef DEBUG1
	debug << "Doing combinatorial mutation"<<flush;
#endif 

	long num_individuals = pool->GetNumIndividuals();
#ifdef DEBUG1
	debug << " GetNumIndividuals() = "<<num_individuals<<flush;
#endif 
	for (int i=0; i<num_individuals; i++)
	{
		long num_chromosomes = pool->GetNumChromosomes(i);
#ifdef DEBUG1
	        debug << " GetNumChromosomes() = "<<num_chromosomes<<flush;
#endif 
		for (int c=0; c<num_chromosomes; c++)
		{
			long num_genes = pool->GetNumGenes(i, c);
#ifdef DEBUG1
	                debug << " GetNumGenes() = "<<num_genes<<" "<<flush;
#endif 
			for (int g=0; g<num_genes; g++)
			{
				if (Flip(GetParameterValue(_rate_param)))
				{
					long site1 = random(num_genes);
					long site2 = site1;

					while (site1 == site2)
						site2 = random(num_genes);

					long value1 = pool->ReadGene(i, c, site1);
					long value2 = pool->ReadGene(i, c, site2);
					pool->WriteGene(i, c, site2, value1);
					pool->WriteGene(i, c, site1, value2);
#ifdef DEBUG1
	                                debug << "f"<<flush;
#endif 
				}
			}
#ifdef DEBUG1
	                debug << endl;
#endif 
		}
	}
}

/**********************************************************************
 MIGRATION OPERATORS
 **********************************************************************/

///////////////////////////////////////////////////////////////////////
// Best-To-Worst Migration
//

BestToWorstMigration::BestToWorstMigration(void)
{
	_ID = BEST_TO_WORST_MIGRATION;
}

void BestToWorstMigration::DoMigration(Pool* src_pool, Pool* dst_pool, OptimiseType min_or_max)
{
	if (!Flip(GetParameterValue(_rate_param)))
        	return;

	#ifdef DEBUG1
	debug << "Doing Migration\n";
	#endif

	long best_index = src_pool->GetHighestFitness(min_or_max); // determine best individual in src pool
	long worst_index = dst_pool->GetLowestFitness(min_or_max); // determine worst individual in dst pool
	
	dst_pool->AddIndividual(worst_index, src_pool->GetIndividualCopy(best_index));
}









