// *******************************************************************************
// File name:		oplib.h
// Author:		Laura Dekker
// Description:		User-defined operators
//
//
// Change history:
//
//
// *******************************************************************************

#ifndef _OPLIB_H
#define _OPLIB_H

#include "qgtypes.h"




///////////////////////////////////////////////////////////////////////
// Operator library data structures
//

struct OperatorLibraryEntry
{
	char _name[LABEL_LENGTH];
	OperatorType _type;
	int _ID;
};


class OperatorLibrary
{
public:
	OperatorLibrary(void);
	~OperatorLibrary(void);

	void CatalogueOperators(void);
	void AddCatalogueEntry(OperatorType type, OperatorID ID, char* name);
	int GetNumOperators(OperatorType type);

	int CreateOperator(OperatorID operatorID);
	BaseOperator* GetOperator(int operator_index);	// instantiated operator
	void DeleteOperator(int operator_index);		// instantiated operator

	char* GetOperatorName(OperatorType type, int cat_index);	// non-instantiated operator
	OperatorID GetOperatorID(OperatorType type, int cat_index);	// non-instantiated operator
	int GetOperatorIndex(OperatorType type, OperatorID ID);		// non-instantiated operator
	int GetOperatorIndex(OperatorType type, char* name);		// non-instantiated operator
	

protected:
	int GetStartIndex(OperatorType type);

	OperatorLibraryEntry* _catalogue[MAX_OPERATORS];
	BaseOperator* _instantiated_operators[MAX_OPERATORS];

	int _num_catalogued;

	int _num_selection;
	int _num_crossover;
	int _num_mutation;
	int _num_migration;
	int _num_initialisation;
	int _num_hillclimber;
	int _num_misc;
};



/**************************************************************************
 Operator parameter class
 **************************************************************************/

struct Parameter
{
	char _description[LABEL_LENGTH];
	ParameterType _type;
	double _value;	
};




/**************************************************************************
 Base operator classes
 **************************************************************************/

class BaseOperator
{
public:
	BaseOperator(void);
	virtual ~BaseOperator(void);

	void SetOperatorName(char* name);
	char* GetOperatorName(void);
	void SetOperatorType(OperatorType);
	OperatorType GetOperatorType(void);	
	OperatorID GetOperatorID(void);

	int AddNumericParameter(char* description, ParameterType type, double default_value = 0);
	int AddStringParameter(char* description, ParameterType type, char* default_value = "");
	char GetNumParameters(void);
	char* GetParameterDescription(int index);
	ParameterType GetParameterType(int index);
	double GetParameterValue(int index);
	void SetParameterValue(int index, double value);
	char* GetParameterString(int index);
	void SetParameterString(int index, char* value);

protected:
	OperatorType _type;
	char _name[LABEL_LENGTH];
	OperatorID _ID;

	Parameter _parameter_array[MAX_PARAMS];
	char _num_parameters;
};





class Crossover : public BaseOperator
{
public:
	Crossover(void);
	virtual void DoCrossover(Pool* pool) = 0;
protected:
	int _rate_param;
};


class Mutation : public BaseOperator
{
public:
	Mutation(void);
	virtual void DoMutation(Pool* pool) = 0;
protected:
	int _rate_param;

};


class Selection : public BaseOperator
{
public:
	Selection(void);
	virtual void DoSelection(Pool* pool, OptimiseType min_or_max) = 0;
protected:

};


class Initialisation : public BaseOperator
{
public:
	Initialisation(void);
	virtual void DoInitialisation(Pool* pool) = 0;
protected:

};


class Migration : public BaseOperator
{
public:
	Migration(void);
	virtual void DoMigration(Pool* source, Pool* destination, OptimiseType min_or_max) = 0;
protected:
	int _rate_param;

};


class HillClimbing : public BaseOperator
{
public:
	HillClimbing(void);
	virtual void DoHillClimbing(Individual* individual) = 0;
protected:

};





/**********************************************************************
  OPERATORS
 **********************************************************************/

///////////////////////////////////////////////////////////////////////
// Initialisation Operators
//

class RandomInitialise : public Initialisation
{
public:
	RandomInitialise(void);
	void DoInitialisation(Pool* pool);
};

class InitFromFile : public Initialisation
{
public:
	InitFromFile(void);
	void DoInitialisation(Pool* pool);
protected:
	int _filename_param;
};


class CombinatorialInitialise : public Initialisation
{
public:
	CombinatorialInitialise(void);
	void DoInitialisation(Pool* pool);
protected:
};

///////////////////////////////////////////////////////////////////////
// Selection Operators
//

class TruncatedSelection : public Selection
{
public:
	TruncatedSelection(void);
	void DoSelection(Pool* pool, OptimiseType min_or_max);
protected:
	int _truncate_param;
	int _norm_scale_param;
};

class TruncatedSelection2 : public Selection
{
public:
	TruncatedSelection2(void);
	void DoSelection(Pool* pool, OptimiseType min_or_max);
protected:
	int _truncate_param;
};

class RouletteWheel : public Selection
{
public:
	RouletteWheel(void);
	void DoSelection(Pool* pool, OptimiseType min_or_max);
protected:
	int _norm_scale_param;
	int _elitism_flag_param;
};

class TournamentSelection : public Selection
{
public:
	TournamentSelection(void);
	void DoSelection(Pool* pool, OptimiseType min_or_max);
protected:
	int _grp_sz_param;
};

class LinearSelection : public Selection
{
public:
	LinearSelection(void);
	void DoSelection(Pool* pool, OptimiseType min_or_max);
protected:
	int _constant_param;
	int _gradient_param;
};

class RemainderStochastic : public Selection
{
public:
	RemainderStochastic(void);
	void DoSelection(Pool* pool, OptimiseType min_or_max);
};

class WindowSelection : public Selection
{
public:
	WindowSelection(void);
	void DoSelection(Pool* pool, OptimiseType min_or_max);
};

///////////////////////////////////////////////////////////////////////
// Crossover Operators
//

class HollandCrossover : public Crossover
{
public:
	HollandCrossover(void);
	void DoCrossover(Pool* pool);
};

class TwoPointCrossover : public Crossover
{
public:
	TwoPointCrossover(void);
	void DoCrossover(Pool* pool);
};

class UniformCrossover : public Crossover
{
public:
	UniformCrossover(void);
	void DoCrossover(Pool* pool);
};


class CombinatorialCrossover : public Crossover
{
public:
	CombinatorialCrossover(void);
	void DoCrossover(Pool* pool);
};


///////////////////////////////////////////////////////////////////////
// Mutation Operators
//

class BasicMutation : public Mutation
{
public:
	BasicMutation(void);
	void DoMutation(Pool* pool);
};

class CombinatorialMutation : public Mutation
{
public:
	CombinatorialMutation(void);
	void DoMutation(Pool* pool);
};


///////////////////////////////////////////////////////////////////////
// Migration Operators
//

class BestToWorstMigration : public Migration
{
public:
	BestToWorstMigration(void);
	void DoMigration(Pool* source, Pool* destination, OptimiseType min_or_max);
};



#endif