// *******************************************************************************
// File name:		pools.cpp
// Author:		Laura Dekker
// Description:		Pools class
//
//
// Change history:
//
//
// *******************************************************************************

#include "pools.h"
#include "algorthm.h"
#include "rand.h"
#include "evaluate.h" 

#define DEBUG

/*****************************************************************
 Class Pool
 *****************************************************************/

Pool::Pool(void)
{
	_num_individuals = 0;
	_individual_array = 0;

	_order_array = 0;
	_order_valid = FALSE;

	_overall_best = 0;
	_old_avg_fitness = 0;
}


Pool::Pool(long num_individuals)
{
	_num_individuals = num_individuals;

	// create the array for individuals, but don't put anything in
	_individual_array = new Individual*[_num_individuals];
	for (long i=0; i<_num_individuals; i++)
		_individual_array[i] = 0;

	_order_array = 0;
	_order_valid = FALSE;

	_overall_best = 0;
	_old_avg_fitness = 0;
}



Pool::Pool(long num_individuals, long num_chromosomes, long num_genes, GeneType type)
{
	_num_individuals = num_individuals;

	_individual_array = new Individual*[_num_individuals];
	for (long i=0; i<_num_individuals; i++)
        {
		_individual_array[i] = new Individual(num_chromosomes, num_genes, type);
		_individual_array[i]->_pool_index = i;
	}

	_order_array = 0;
	_order_valid = FALSE;

	_overall_best = 0;
	_old_avg_fitness = 0;
}


Pool::Pool(long num_individuals, Individual* prototype)
{
	_num_individuals = num_individuals;

	_individual_array = new Individual*[_num_individuals];
	for (long i=0; i<_num_individuals; i++)
        {
		_individual_array[i] = new Individual(*prototype);
		_individual_array[i]->_pool_index = i;
	}

	_order_array = 0;
	_order_valid = FALSE;

	_overall_best = 0;
	_old_avg_fitness = 0;

}


Pool::Pool(const Pool& old_pool)
{
	_num_individuals = 0;
	_individual_array = 0;
	_order_array = 0;
	_overall_best = 0;

	*this = old_pool;	
}


Pool& Pool::operator=(const Pool& rhs)
{
	for (long i=0; i<_num_individuals; i++)       
		delete _individual_array[i];
       
	delete[] _individual_array;
	delete[] _order_array;

	_num_individuals = rhs._num_individuals;

	_individual_array = new Individual*[_num_individuals];
	_order_array = new long[_num_individuals];

	for (i=0; i<_num_individuals; i++)
	{
		_individual_array[i] = new Individual(*(rhs._individual_array[i]));
		_order_array[i] = rhs._order_array[i];
	}

	_order_valid = rhs._order_valid;
	
	_old_avg_fitness = rhs._old_avg_fitness;
	_overall_best = rhs._overall_best;

	return *this;
}


Pool::~Pool(void)
{
	for (long i=0; i<_num_individuals; i++)
		delete _individual_array[i];
	delete[] _individual_array;
	delete[] _order_array;
	delete _overall_best;
} 




/*****************************************************************
 iostreams
 *****************************************************************/

ostream& operator<< (ostream& os, const Pool& pool)
{
	for (long i=0; i<pool._num_individuals; i++)
        {
		os << *(pool._individual_array[i]) << '\n';
	}
	os << '\n';
	return os;
}


Pool& operator>> (istream& is, Pool& pool)
{
	for (long i=0; i<pool._num_individuals; i++)
		is >> *(pool._individual_array[i]);

	return pool;
} 



/*****************************************************************
 Actions on individuals
 *****************************************************************/


long Pool::GetNumIndividuals(void)
{
	return _num_individuals;
}


Individual* Pool::GetIndividualHandle(long index)
{
	if (index < _num_individuals)
		return _individual_array[index];
	else
		return 0;
}



Individual* Pool::GetIndividualCopy(long index)
{
	if (index < _num_individuals)        
		return new Individual(*(_individual_array[index]));
	else
		return 0;
}


void Pool::AddIndividual(long index, Individual* individual)
{
	if (_num_individuals >= index)
	{
		delete _individual_array[index];
		_individual_array[index] = individual;
		_individual_array[index]->_pool_index = index;
	}
	_order_valid = FALSE;
}


void Pool::DeleteIndividual(long index)
{
	if (_num_individuals >= index)
	{
		delete _individual_array[index];
		_individual_array[index] = 0;
	}

	_order_valid = FALSE;
}



Individual* Pool::RemoveIndividual(long index)
{
	if (index >= _num_individuals)
		return 0;
	
	Individual* returning = _individual_array[index];
	returning->_pool_index = INVALID_INDEX;
	_individual_array[index] = 0;

	_order_valid = FALSE;

	return returning;
}



/*****************************************************************
 Actions on chromosomes
 *****************************************************************/

long Pool::GetNumChromosomes(long individual)
{
	return _individual_array[individual]->_num_chromosomes;
}


Chromosome* Pool::GetChromosomeHandle(long individual, long chromosome)
{
	return _individual_array[individual]->_chromosome_array[chromosome];
}


Chromosome* Pool::GetChromosomeCopy(long individual, long chromosome)
{
	return new Chromosome(*(_individual_array[individual]->_chromosome_array[chromosome]));
}


void Pool::AddChromosome(long individual, long chromosome, Chromosome* new_chromosome)
{
	delete _individual_array[individual]->_chromosome_array[chromosome];
	_individual_array[individual]->_chromosome_array[chromosome] = new_chromosome;

	_individual_array[individual]->_fitness_valid = FALSE;
	_order_valid = FALSE;
}


void Pool::DeleteChromosome(long individual, long chromosome)
{
	delete _individual_array[individual]->_chromosome_array[chromosome];
	_individual_array[individual]->_chromosome_array[chromosome] = 0;

	_individual_array[individual]->_fitness_valid = FALSE;
	_order_valid = FALSE;
}



/*****************************************************************
 Actions on genes
 *****************************************************************/

long Pool::GetNumGenes(long individual, long chromosome)
{
	return _individual_array[individual]->_chromosome_array[chromosome]->_num_genes;
}


double Pool::ReadGene(long individual, long chromosome, long gene)
{
	double value;

	if (individual < _num_individuals
		&& chromosome < _individual_array[individual]->_num_chromosomes
		&& gene < _individual_array[individual]->_chromosome_array[chromosome]->_num_genes)
	{
		switch (_individual_array[individual]->_chromosome_array[chromosome]->_type)
		{
			case G_ENCODED:
				value = ((char*)(_individual_array[individual]->_chromosome_array[chromosome]->_gene_array))[gene];
				break;

			case G_INT:
				value = ((long*)(_individual_array[individual]->_chromosome_array[chromosome]->_gene_array))[gene];
				break;

			case G_FLOAT:
				value = ((double*)(_individual_array[individual]->_chromosome_array[chromosome]->_gene_array))[gene];
				break;
		}
	}
	return value;
}


void Pool::WriteGene(long individual, long chromosome, long gene, double value)
{
	Chromosome* this_chromosome;

	// check that the given indices are OK
	if (individual < _num_individuals
		&& chromosome < _individual_array[individual]->_num_chromosomes
		&& gene < _individual_array[individual]->_chromosome_array[chromosome]->_num_genes)
	{
		this_chromosome = _individual_array[individual]->_chromosome_array[chromosome];

		if (this_chromosome->_type == G_ENCODED)
		{
			((char*)(this_chromosome->_gene_array))[gene] = value;
			// if we've just written a bit that takes the decoded
			// value out of range, truncate it
			if (this_chromosome->BaseDecode() > this_chromosome->_max)
				this_chromosome->BaseEncode(this_chromosome->_max);
		}

		else
                {
			// if gene value is out of bounds, truncate to min/max		
			if (value < this_chromosome->_min)
				value = this_chromosome->_min;
			else if (value > this_chromosome->_max)
				value = this_chromosome->_max;

			switch (this_chromosome->_type)
			{
				case G_INT:
					((long*)(this_chromosome->_gene_array))[gene] = value;
					break;
                                              
				case G_FLOAT:
					((double*)(this_chromosome->_gene_array))[gene] = value;
					break;
			}
		}	

		// new value, so fitness cache is invalid and order cache is invalid
		_individual_array[individual]->_fitness_valid = FALSE;
		_order_valid = FALSE;
	}
}


// set the min for the whole chromosome
void Pool::SetMinValue(long individual, long chromosome, double min)
{	_individual_array[individual]->_chromosome_array[chromosome]->_min = min;	}

// set the min for a particular gene
void Pool::SetMinValue(long individual, long chromosome, long gene, double min)
{
	if (_individual_array[individual]->_chromosome_array[chromosome]->_type == G_USER)
		((Gene*)(_individual_array[individual]->_chromosome_array[chromosome]->_gene_array))[gene]._min = min;
	else
		_individual_array[individual]->_chromosome_array[chromosome]->_min = min;
}

double Pool::GetMinValue(long individual, long chromosome)
{	return _individual_array[individual]->_chromosome_array[chromosome]->_min;	}

double Pool::GetMinValue(long individual, long chromosome, long gene)
{
	if (_individual_array[individual]->_chromosome_array[chromosome]->_type == G_USER)
		return ((Gene*)(_individual_array[individual]->_chromosome_array[chromosome]->_gene_array))[gene]._min;
	else
		return _individual_array[individual]->_chromosome_array[chromosome]->_min;
}

void Pool::SetMaxValue(long individual, long chromosome, double max)
{	_individual_array[individual]->_chromosome_array[chromosome]->_max = max;	}

void Pool::SetMaxValue(long individual, long chromosome, long gene, double max)
{
	if (_individual_array[individual]->_chromosome_array[chromosome]->_type == G_USER)
		((Gene*)(_individual_array[individual]->_chromosome_array[chromosome]->_gene_array))[gene]._max = max;
	else
		_individual_array[individual]->_chromosome_array[chromosome]->_max = max;
}

double Pool::GetMaxValue(long individual, long chromosome)
{	return _individual_array[individual]->_chromosome_array[chromosome]->_max;	}

double Pool::GetMaxValue(long individual, long chromosome, long gene)
{
	if (_individual_array[individual]->_chromosome_array[chromosome]->_type == G_USER)
		return ((Gene*)(_individual_array[individual]->_chromosome_array[chromosome]->_gene_array))[gene]._max;
	else
		return _individual_array[individual]->_chromosome_array[chromosome]->_max;
}

GeneType Pool::GetGeneType(long individual, long chromosome)
{	return _individual_array[individual]->_chromosome_array[chromosome]->_type;	}




char Pool::GetAlphabet(long individual, long chromosome)
{
	if (_individual_array[individual]->_chromosome_array[chromosome]->_type == G_ENCODED)
		return _individual_array[individual]->_chromosome_array[chromosome]->_alphabet;
	else
		return 0;
}




/*****************************************************************
 Fitness evaluation, etc.
 *****************************************************************/

FitnessType Pool::GetFitness(long index)
{
	return _individual_array[index]->GetFitness();
}


void Pool::InvalidateFitness(long index)
{
	_individual_array[index]->_fitness_valid = FALSE;
	_order_valid = FALSE;
}


FitnessType Pool::GetTotalFitness(void)
{
	FitnessType sum_fitness = 0;

	for (long i=0; i<_num_individuals; i++)
		sum_fitness += GetFitness(i);

	return sum_fitness;
}


FitnessType Pool::GetAverageFitness(void)
{
	return GetTotalFitness()/_num_individuals;
}


long Pool::GetHighestFitness(OptimiseType min_or_max)
{
	if (!_order_valid)        
		SortPool();

	if (min_or_max == MAXIMISE)
		return _order_array[0];
	else
		return _order_array[_num_individuals-1];
}


long Pool::GetLowestFitness(OptimiseType min_or_max)
{
	if (!_order_valid)
		SortPool();

	if (min_or_max == MAXIMISE)
		return _order_array[_num_individuals-1];
	else
		return _order_array[0];
}


long Pool::GetNthHighest(long nth, OptimiseType min_or_max)
{
	if (nth < _num_individuals)
	{
		if (!_order_valid)
			SortPool();

		if (min_or_max == MAXIMISE)
			return _order_array[nth];
		else
			return _order_array[_num_individuals-(nth+1)];
	}
	else
		return INVALID_INDEX;
}



void Pool::SetOldAvgFitness(FitnessType old_avg_fitness)
{	_old_avg_fitness = old_avg_fitness;	}



FitnessType Pool::GetOldAvgFitness(void)
{	return _old_avg_fitness;	}



void Pool::StoreOverallBest(OptimiseType min_or_max)
{
	Individual* current_best;
	long current_best_index;
	FitnessType overall_best_fitness;

	if (!_overall_best)	// if no overall best has been stored yet
	{
		current_best_index = GetHighestFitness(min_or_max);
		_overall_best = GetIndividualCopy(current_best_index);
	}

	else
        { 
		overall_best_fitness = _overall_best->GetFitness();
		current_best_index = GetHighestFitness(min_or_max);
		current_best = GetIndividualHandle(current_best_index);

		if ((min_or_max == MINIMISE  &&  current_best->GetFitness() < overall_best_fitness)
			|| (min_or_max == MAXIMISE  &&  current_best->GetFitness() > overall_best_fitness))
		{
			delete _overall_best;
			_overall_best = GetIndividualCopy(current_best_index);
		}
        }
}  



Individual* Pool::GetOverallBest(void)
{	return _overall_best;	}



/*****************************************************************
 Sort the pool in terms of fitness values
 This DOES NOT sort the pool itself - just stores an array
 indicating their order of rank 
 *****************************************************************/

void Pool::SortPool(void)
{
	FitnessType* fitness = new FitnessType[_num_individuals];
	FitnessType temp=0;
	int changed;

	if (!_order_array)
		_order_array = new long[_num_individuals];

	// fill fitness array and order array
	for (long i=0; i<_num_individuals; i++)
	{
		fitness[i] = GetFitness(i);     
		_order_array[i] = i;
	}

	changed=FALSE;
	i=0;

	// loop until no changes are made in order array
	while(i<_num_individuals-1 || changed==TRUE)
	{
		i=0;
		changed=FALSE;
		// loop through arrays, swapping if first element is less than second
		while(i<_num_individuals-1)
		{
			if (fitness[i]<fitness[i+1])
			{
				temp=fitness[i];
				fitness[i]=fitness[i+1];
				fitness[i+1]=temp;
				temp=_order_array[i];
				_order_array[i]=_order_array[i+1];
				_order_array[i+1]=temp;
				changed=TRUE;                           
			}                       
			i++;
		}
	}
	delete[] fitness;

	_order_valid = TRUE;
}


/*****************************************************************
 Jumble pool.  Rearranges the individuals in the pool into a 
 random order.
 *****************************************************************/

void Pool::JumblePool(void)
{ 	
	long* indices = new long[_num_individuals];
	Individual** jumble_array = new Individual*[_num_individuals];

	long index;

	const int deleted = -1;

	for (long i=0; i<_num_individuals; i++)
		indices[i]=i;

	for (i=0; i<_num_individuals; i++)
	{ 		               
		index = Int_Rand(0,_num_individuals-1);
		while (indices[index] == deleted)		
			index = (index+1)%_num_individuals;
		jumble_array[i] = _individual_array[index];
		indices[index] = deleted;
	}

	for (i=0; i<_num_individuals; i++)
        {
		_individual_array[i] = jumble_array[i];
		_individual_array[i]->_pool_index = i;
        }

	_order_valid = FALSE;
	delete[] indices;
	delete[] jumble_array;
}



/*****************************************************************
 Statistics, etc.
 *****************************************************************/

ostream& Pool::Statistics(ostream& os)
{
	os << "\nGenotypes:\n";
	for (long i=0; i<_num_individuals; i++)
	{	
		GetIndividualHandle(i)->PrintGenotype(os);
		os << '\n';
	}

	os << "\nPhenotypes and fitness values:\n";
	for (i=0; i<_num_individuals; i++)
	{
		GetIndividualHandle(i)->PrintPhenotype(os);
		os << '\n';
	}

	return os;
}


ostream& Pool::Fitnesses(ostream& os)
{
	for (long i=0; i<_num_individuals; i++)
	{
		FitnessType fitness = GetIndividualHandle(i)->GetFitness();
		os << fitness << '\t';
	}

	return os;
}





/*****************************************************************
 Global functions
 *****************************************************************/


void CopyIndividual(Pool* src, long si, Pool* dst, long di)
{
	Individual* copy = src->GetIndividualCopy(si);
	dst->AddIndividual(di, copy);
}



void SwapGenes (Pool* src, long si, long sc,
				Pool* dst, long di, long dc,
				long swap_index, long num_genes)
{
	double swap_value;

	// check that the chromosomes are long enough to do a swap at this index
    // >>> assuming both chromosomes of the same length <<<
	long chrom_length = src->GetNumGenes(si, sc);

	if (chrom_length>swap_index)
	{
		// check chromosome limits
		if ((num_genes == MAX_GENES) || ((swap_index+num_genes) >= chrom_length))
    		num_genes = chrom_length - swap_index;

		for (long g=swap_index; num_genes > 0; g++, num_genes--)
		{
			swap_value = src->ReadGene(si, sc, g);
			src->WriteGene(si, sc, g, dst->ReadGene(di,dc,g));
			dst->WriteGene(di, dc, g, swap_value);
		}
	}
	src->InvalidateFitness(si);
	dst->InvalidateFitness(di);
}





